@@ -38,8 +38,6 @@
  * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
  * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
  * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
- *
- * $Id: lmtpd.c,v 1.167 2010/05/25 20:59:19 wescraig Exp $
  */
 
 #include <config.h>
@@ -90,6 +88,7 @@
 #include "statuscache.h"
 #include "telemetry.h"
 #include "tls.h"
+#include "userdeny.h"
 #include "util.h"
 #include "version.h"
 #include "xmalloc.h"
@@ -133,9 +132,10 @@ const int config_need_data = CONFIG_NEED_PARTITION_DATA;
 
 extern int optind;
 extern char *optarg;
-static int dupelim = 1;		/* eliminate duplicate messages with
-				   same message-id */
-static int singleinstance = 1;	/* attempt single instance store */
+static int dupelim = 1;         /* eliminate duplicate messages with
+                                   same message-id */
+static int singleinstance = 1;  /* attempt single instance store */
+static int isproxy = 0;
 
 struct stagemsg *stage = NULL;
 
@@ -144,7 +144,6 @@ static struct protstream *deliver_out, *deliver_in;
 int deliver_logfd = -1; /* used in lmtpengine.c */
 
 /* our cached connections */
-mupdate_handle *mhandle = NULL;
 struct backend **backend_cached = NULL;
 
 static struct protocol_t lmtp_protocol =
@@ -164,9 +163,9 @@ static struct protocol_t lmtp_protocol =
 };
 
 static struct sasl_callback mysasl_cb[] = {
-    { SASL_CB_GETOPT, &mysasl_config, NULL },
-    { SASL_CB_PROXY_POLICY, &mysasl_proxy_policy, NULL },
-    { SASL_CB_CANON_USER, &mysasl_canon_user, NULL },
+    { SASL_CB_GETOPT, (mysasl_cb_ft *) &mysasl_config, NULL },
+    { SASL_CB_PROXY_POLICY, (mysasl_cb_ft *) &mysasl_proxy_policy, NULL },
+    { SASL_CB_CANON_USER, (mysasl_cb_ft *) &mysasl_canon_user, NULL },
     { SASL_CB_LIST_END, NULL, NULL }
 };
 
@@ -186,16 +185,15 @@ int service_init(int argc __attribute__((unused)),
 
     global_sasl_init(1, 1, mysasl_cb);
 
+    /* so we can do mboxlist operations */
+    mboxlist_init(0);
+    mboxlist_open(NULL);
+
     if (config_mupdate_server &&
-	(config_mupdate_config == IMAP_ENUM_MUPDATE_CONFIG_STANDARD) &&
-	!config_getstring(IMAPOPT_PROXYSERVERS)) {
-	/* proxy only -- talk directly to mupdate master */
-	r = mupdate_connect(config_mupdate_server, NULL, &mhandle, NULL);
-	if (r) {
-	    syslog(LOG_ERR, "couldn't connect to MUPDATE server %s: %s",
-		   config_mupdate_server, error_message(r));
-	    fatal("error connecting with MUPDATE server", EC_TEMPFAIL);
-	}
+        (config_mupdate_config == IMAP_ENUM_MUPDATE_CONFIG_STANDARD) &&
+        !config_getstring(IMAPOPT_PROXYSERVERS)) {
+        /* proxy only */
+        isproxy = 1;
     }
     else {
 	dupelim = config_getswitch(IMAPOPT_DUPLICATESUPPRESSION);
@@ -218,14 +216,14 @@ int service_init(int argc __attribute__((unused)),
 	    }
 	}
 
-	/* so we can do mboxlist operations */
-	mboxlist_init(0);
-	mboxlist_open(NULL);
-
 	/* so we can do quota operations */
 	quotadb_init(0);
 	quotadb_open(NULL);
 
+	/* open the user deny db */
+	denydb_init(0);
+	denydb_open(NULL);
+
 	/* Initialize the annotatemore db (for sieve on shared mailboxes) */
 	annotatemore_init(0, NULL, NULL);
 	annotatemore_open(NULL);
@@ -234,7 +232,7 @@ int service_init(int argc __attribute__((unused)),
 	statuscache_open(NULL);
 
 	/* setup for sending IMAP IDLE notifications */
-	idle_enabled();
+	idle_init();
     }
 
     /* Set namespace */
@@ -250,24 +248,13 @@ int service_init(int argc __attribute__((unused)),
     return 0;
 }
 
-static int mupdate_ignore_cb(struct mupdate_mailboxdata *mdata __attribute__((unused)),
-			     const char *cmd __attribute__((unused)),
-			     void *context __attribute__((unused))) 
-{
-    /* If we get called, we've recieved something other than an OK in
-     * response to the NOOP, so we want to hang up this connection anyway */
-    return MUPDATE_FAIL;
-}
-
 /*
  * run for each accepted connection
  */
 int service_main(int argc, char **argv, 
 		 char **envp __attribute__((unused)))
 {
-    int opt, r;
-
-    session_new_id();
+    int opt;
 
     sync_log_init();
 
@@ -290,32 +277,7 @@ int service_main(int argc, char **argv,
     snmp_increment(TOTAL_CONNECTIONS, 1);
     snmp_increment(ACTIVE_CONNECTIONS, 1);
 
-    /* get a connection to the mupdate server */
-    r = 0;
-    if (mhandle) {
-	/* we have one already, test it */
-	r = mupdate_noop(mhandle, mupdate_ignore_cb, NULL);
-	if (r) {
-	    /* will NULL mhandle for us */
-	    mupdate_disconnect(&mhandle);
-
-	    /* connect to the mupdate server */
-	    r = mupdate_connect(config_mupdate_server, NULL, &mhandle, NULL);
-	}
-    }
-    if (!r) {
-	lmtpmode(&mylmtp, deliver_in, deliver_out, 0);
-    } else {
-	syslog(LOG_ERR, "couldn't connect to %s: %s", config_mupdate_server,
-	       error_message(r));
-	prot_printf(deliver_out, "451");
-	if (config_serverinfo) prot_printf(deliver_out, " %s", config_servername);
-	if (config_serverinfo == IMAP_ENUM_SERVERINFO_ON) {
-	    prot_printf(deliver_out, " Cyrus LMTP%s %s",
-			config_mupdate_server ? " Murder" : "", cyrus_version());
-	}
-	prot_printf(deliver_out, " %s\r\n", error_message(r));
-    }
+    lmtpmode(&mylmtp, deliver_in, deliver_out, 0);
 
     /* free session state */
     if (deliver_in) prot_free(deliver_in);
@@ -338,8 +300,7 @@ void service_abort(int error)
     shut_down(error);
 }
 
-static void
-usage()
+static void usage()
 {
     fprintf(stderr, "421-4.3.0 usage: lmtpd [-C <alt_config>] [-a]\r\n");
     fprintf(stderr, "421 4.3.0 %s\n", cyrus_version());
@@ -429,55 +390,32 @@ static int mlookup(const char *name, char **server, char **aclp, void *tid)
 {
     int r;
     char *c;
+    struct mboxlist_entry mbentry;
 
-    if (server) *server = NULL;
-
-    if (mhandle) {
-	/* proxy only, so check the mupdate master */
-	struct mupdate_mailboxdata *mailboxdata;
-
-	/* find what server we're sending this to */
-	r = mupdate_find(mhandle, name, &mailboxdata);
-
-	if (r == MUPDATE_MAILBOX_UNKNOWN) {
-	    return IMAP_MAILBOX_NONEXISTENT;
-	} else if (r) {
-	    /* xxx -- yuck: our error handling for now will be to exit;
-	       this txn will be retried later -- to do otherwise means
-	       that we may have to restart this transaction from scratch */
-	    fatal("error communicating with MUPDATE server", EC_TEMPFAIL);
-	}
-
-	if (aclp) *aclp = (char *) mailboxdata->acl;
-	if (server) *server = (char *) mailboxdata->server;
-	c = strchr(*server, '!');
-	if (c) *c = '\0';
+    /* do a local lookup and kick the slave if necessary */
+    r = mboxlist_lookup(name, &mbentry, tid);
+    if (r == IMAP_MAILBOX_NONEXISTENT && config_mupdate_server) {
+        kick_mupdate();
+        r = mboxlist_lookup(name, &mbentry, tid);
     }
-    else {
-	struct mboxlist_entry mbentry;
-	/* do a local lookup and kick the slave if necessary */
-	r = mboxlist_lookup(name, &mbentry, tid);
-	if (r == IMAP_MAILBOX_NONEXISTENT && config_mupdate_server) {
-	    kick_mupdate();
-	    r = mboxlist_lookup(name, &mbentry, tid);
-	}
-	if (r) return r;
+    if (r) return r;
+    if (mbentry.mbtype & MBTYPE_MOVING)  return IMAP_MAILBOX_MOVED;
+    if (mbentry.mbtype & MBTYPE_DELETED) return IMAP_MAILBOX_NONEXISTENT;
 
-	if (aclp) *aclp = mbentry.acl;
-	if (server) {
-	    if (mbentry.mbtype & MBTYPE_REMOTE) {
-		/* xxx hide the fact that we are storing partitions */
-		*server = mbentry.partition;
-		c = strchr(*server, '!');
-		if (c) *c = '\0';
-	    }
-	}
+    if (aclp) *aclp = mbentry.acl;
+    if (server) {
+        if (mbentry.mbtype & MBTYPE_REMOTE) {
+            /* xxx hide the fact that we are storing partitions */
+            *server = mbentry.partition;
+            c = strchr(*server, '!');
+            if (c) *c = '\0';
+        }
     }
 
     return r;
 }
 
-/* places msg in mailbox mailboxname.  
+/* places msg in mailbox mailboxname.
  * if you wish to use single instance store, pass stage as non-NULL
  * if you want to deliver message regardless of duplicates, pass id as NULL
  * if you want to notify, pass user
@@ -488,14 +426,14 @@ int deliver_mailbox(FILE *f,
 		    struct message_content *content,
 		    struct stagemsg *stage,
 		    unsigned size,
-		    char **flag,
-		    int nflags,
+		    const strarray_t *flags,
 		    char *authuser,
 		    struct auth_state *authstate,
 		    char *id,
 		    const char *user,
 		    char *notifyheader,
 		    const char *mailboxname,
+		    char *date,
 		    int quotaoverride,
 		    int acloverride)
 {
@@ -503,6 +441,7 @@ int deliver_mailbox(FILE *f,
     struct appendstate as;
     unsigned long uid;
     const char *notifier;
+    duplicate_key_t dkey = {NULL, NULL, NULL};
 
     r = append_setup(&as, mailboxname,
 		     authuser, authstate, acloverride ? 0 : ACL_POST, 
@@ -511,9 +450,12 @@ int deliver_mailbox(FILE *f,
 		     (long) size : 0);
 
     /* check for duplicate message */
+    dkey.id = id;
+    dkey.to = mailboxname;
+    dkey.date = date;
     if (!r && id && dupelim && !(as.mailbox->i.options & OPT_IMAP_DUPDELIVER) &&
-	duplicate_check(id, strlen(id), mailboxname, strlen(mailboxname))) {
-	duplicate_log(id, mailboxname, "delivery");
+	duplicate_check(&dkey)) {
+	duplicate_log(&dkey, "delivery");
 	append_abort(&as);
 	return 0;
     }
@@ -526,7 +468,7 @@ int deliver_mailbox(FILE *f,
 
     if (!r) {
 	r = append_fromstage(&as, &content->body, stage, 0,
-			     (const char **) flag, nflags, !singleinstance);
+			     flags, !singleinstance);
 
 	if (r) {
 	    append_abort(&as);
@@ -539,8 +481,7 @@ int deliver_mailbox(FILE *f,
 		syslog(LOG_INFO, "Delivered: %s to mailbox: %s",
 		       id, mailboxname);
 		if (dupelim && id) {
-		    duplicate_mark(id, strlen(id), mailboxname, 
-				   strlen(mailboxname), time(NULL), uid);
+		    duplicate_mark(&dkey, time(NULL), uid);
 		}
 		mailbox_close(&mailbox);
 	    }
@@ -561,7 +502,7 @@ int deliver_mailbox(FILE *f,
 	    if (strlen(mailboxname) >= inboxlen &&
 		!strncmp(mailboxname, inbox, inboxlen) &&
 		(!mailboxname[inboxlen] || mailboxname[inboxlen] == '.')) {
-		strlcpy(inbox, "INBOX", sizeof(inbox)); 
+		strlcpy(inbox, "INBOX", sizeof(inbox));
 		strlcat(inbox, mailboxname+inboxlen, sizeof(inbox));
 		notify_mailbox = inbox;
 	    }
@@ -644,7 +585,7 @@ void deliver_remote(message_data_t *msgdata,
 	    int j = rc->rcpt_num;
 	    switch (status[j]) {
 	    case s_wait:
-		/* hmmm, if something fails we'll want to try an 
+		/* hmmm, if something fails we'll want to try an
 		   error delivery */
 		if (lt->rcpt[i].result != RCPT_GOOD) {
 		    status[j] = s_err;
@@ -673,7 +614,7 @@ void deliver_remote(message_data_t *msgdata,
     }
 }
 
-int deliver_local(deliver_data_t *mydata, char **flag, int nflags,
+int deliver_local(deliver_data_t *mydata, const strarray_t *flags,
 		  const char *username, const char *mailboxname)
 {
     char namebuf[MAX_MAILBOX_BUFFER] = "", *tail;
@@ -687,10 +628,10 @@ int deliver_local(deliver_data_t *mydata, char **flag, int nflags,
 	strlcat(namebuf, mailboxname, sizeof(namebuf));
 
 	return deliver_mailbox(md->f, mydata->content, mydata->stage,
-			       md->size, flag, nflags,
+			       md->size, flags,
 			       mydata->authuser, mydata->authstate, md->id,
 			       NULL, mydata->notifyheader,
-			       namebuf, quotaoverride, 0);
+			       namebuf, md->date, quotaoverride, 0);
     }
 
     /* case 2: ordinary user */
@@ -707,20 +648,20 @@ int deliver_local(deliver_data_t *mydata, char **flag, int nflags,
 	    strlcat(namebuf, mailboxname, sizeof(namebuf));
 
 	    ret2 = deliver_mailbox(md->f, mydata->content, mydata->stage,
-				   md->size, flag, nflags,
+				   md->size, flags,
 				   mydata->authuser, mydata->authstate, md->id,
 				   username, mydata->notifyheader,
-				   namebuf, quotaoverride, 0);
+				   namebuf, md->date, quotaoverride, 0);
 	}
 	if (ret2 == IMAP_MAILBOX_NONEXISTENT && mailboxname &&
 	    config_getswitch(IMAPOPT_LMTP_FUZZY_MAILBOX_MATCH) &&
 	    fuzzy_match(namebuf)) {
 	    /* try delivery to a fuzzy matched mailbox */
 	    ret2 = deliver_mailbox(md->f, mydata->content, mydata->stage,
-				   md->size, flag, nflags,
+				   md->size, flags,
 				   mydata->authuser, mydata->authstate, md->id,
 				   username, mydata->notifyheader,
-				   namebuf, quotaoverride, 0);
+				   namebuf, md->date, quotaoverride, 0);
 	}
 	if (ret2) {
 	    /* normal delivery to INBOX */
@@ -729,10 +670,10 @@ int deliver_local(deliver_data_t *mydata, char **flag, int nflags,
 	    *tail = '\0';
 
 	    ret = deliver_mailbox(md->f, mydata->content, mydata->stage,
-				  md->size, flag, nflags,
+				  md->size, flags,
 				  (char *) username, authstate, md->id,
 				  username, mydata->notifyheader,
-				  namebuf, quotaoverride, 1);
+				  namebuf, md->date, quotaoverride, 1);
 
 	    if (authstate) auth_freestate(authstate);
 	}
@@ -750,7 +691,7 @@ int deliver(message_data_t *msgdata, char *authuser,
     struct message_content content = { NULL, 0, NULL };
     char *notifyheader;
     deliver_data_t mydata;
-    
+
     assert(msgdata);
     nrcpts = msg_getnumrcpt(msgdata);
     assert(nrcpts);
@@ -768,7 +709,7 @@ int deliver(message_data_t *msgdata, char *authuser,
     mydata.namespace = &lmtpd_namespace;
     mydata.authuser = authuser;
     mydata.authstate = authstate;
-    
+
     /* loop through each recipient, attempting delivery for each */
     for (n = 0; n < nrcpts; n++) {
 	char namebuf[MAX_MAILBOX_BUFFER] = "", *server;
@@ -818,7 +759,7 @@ int deliver(message_data_t *msgdata, char *authuser,
 #endif
 
 	    if (r) {
-		r = deliver_local(&mydata, NULL, 0, userbuf, mailbox);
+		r = deliver_local(&mydata, NULL, userbuf, mailbox);
 	    }
 	}
 
@@ -837,7 +778,7 @@ int deliver(message_data_t *msgdata, char *authuser,
 	while (d) {
 	    struct dest *nextd = d->next;
 	    struct rcpt *rc = d->to;
-   
+
 	    while (rc) {
 		struct rcpt *nextrc = rc->next;
 		free(rc);
@@ -855,7 +796,7 @@ int deliver(message_data_t *msgdata, char *authuser,
 	    case s_err:
 	    case s_done:
 		/* yikes, we haven't implemented sieve ! */
-		syslog(LOG_CRIT, 
+		syslog(LOG_CRIT,
 		       "sieve states reached, but we don't implement sieve");
 		abort();
 	    break;
@@ -879,7 +820,7 @@ int deliver(message_data_t *msgdata, char *authuser,
 	    assert(status[n] == done || status[n] == s_done);
 	}
     }
-   
+
     /* cleanup */
     free(status);
     if (content.base) map_free(&content.base, &content.len);
@@ -897,7 +838,7 @@ int deliver(message_data_t *msgdata, char *authuser,
 void fatal(const char* s, int code)
 {
     static int recurse_code = 0;
-    
+
     if(recurse_code) {
 	/* We were called recursively. Just give up */
 	snmp_increment(ACTIVE_CONNECTIONS, -1);
@@ -911,7 +852,7 @@ void fatal(const char* s, int code)
     if (stage) append_removestage(stage);
 
     syslog(LOG_ERR, "FATAL: %s", s);
-    
+
     /* shouldn't return */
     shut_down(code);
 
@@ -938,9 +879,10 @@ void shut_down(int code)
     }
     if (backend_cached) free(backend_cached);
 
-    if (mhandle) {
-	mupdate_disconnect(&mhandle);
-    } else {
+    mboxlist_close();
+    mboxlist_done();
+
+    if (!isproxy) {
 #ifdef USE_SIEVE
 	sieve_interp_free(&sieve_interp);
 #else
@@ -948,17 +890,19 @@ void shut_down(int code)
 #endif
 	    duplicate_done();
 
-	mboxlist_close();
-	mboxlist_done();
+        quotadb_close();
+        quotadb_done();
 
-	quotadb_close();
-	quotadb_done();
+	denydb_close();
+	denydb_done();
 
 	annotatemore_close();
 	annotatemore_done();
 
 	statuscache_close();
 	statuscache_done();
+
+	idle_done();
     }
 
 #ifdef HAVE_SSL
@@ -1050,6 +994,18 @@ static int verify_user(const char *user, const char *domain, char *mailbox,
 	}
     }
 
+    if (!r && user) {
+	char msg[MAX_MAILBOX_PATH+1];
+
+	if (domain) {
+	    snprintf(namebuf, sizeof(namebuf), "%s@%s", user, domain);
+	    user = namebuf;
+	}
+
+	if (userdeny(user, config_ident, msg, sizeof(msg)))
+	    return IMAP_MAILBOX_DISABLED;
+    }
+
     if (r) syslog(LOG_DEBUG, "verify_user(%s) failed: %s", namebuf,
 		  error_message(r));
 
@@ -1080,7 +1036,7 @@ char *generate_notify(message_data_t *m)
 		    ret = xrealloc(ret, len += 1024);
 		}
 		pos += sprintf(ret + pos, "%s: ", h);
-		
+
 		/* put the header body.
 		   xxx it would be nice to linewrap.*/
 		/* need: length + '\n' + '\0' */
@@ -1103,7 +1059,7 @@ FILE *spoolfile(message_data_t *msgdata)
 
     /* spool to the stage of one of the recipients
        (don't bother if we're only a proxy) */
-    n = mhandle ? 0 : msg_getnumrcpt(msgdata);
+    n = isproxy ? 0 : msg_getnumrcpt(msgdata);
     for (i = 0; !f && (i < n); i++) {
 	char namebuf[MAX_MAILBOX_BUFFER] = "", *server;
 	const char *user, *domain, *mailbox;
