@@ -56,7 +56,6 @@
 #include <fcntl.h>
 #include <string.h>
 
-#include "md5global.h"
 #include "md5.h"
 #include "sieve_interface.h"
 #include "interp.h"
@@ -110,28 +109,46 @@ int do_reject(action_list_t *a, const char *msg)
  * incompatible with: reject
  */
 int do_fileinto(action_list_t *a, const char *mbox, int cancel_keep,
-		sieve_imapflags_t *imapflags)
+		strarray_t *imapflags)
 {
     action_list_t *b = NULL;
 
     /* see if this conflicts with any previous actions taken on this message */
     while (a != NULL) {
-	b = a;
 	if (a->a == ACTION_REJECT)
 	    return SIEVE_RUN_ERROR;
+	if (a->a == ACTION_FILEINTO && !strcmp(a->u.fil.mailbox, mbox)) {
+	    /* don't bother doing it twice */
+	    /* check that we have a valid action */
+	    if (b == NULL) {
+		return SIEVE_INTERNAL_ERROR;
+	    }
+	    /* cut this action out of the list */
+	    b->next = a->next;
+	    a->next = NULL;
+	    /* find the end of the list */
+	    while (b->next != NULL) {
+		b = b-> next;
+	    }
+	    /* add the action to the end of the list */
+	    b->next = a;
+	    break;
+	}
+	b = a;
 	a = a->next;
     }
 
-    /* add to the action list */
-    a = (action_list_t *) xmalloc(sizeof(action_list_t));
-    if (a == NULL)
-	return SIEVE_NOMEM;
+    if (a == NULL) {
+	/* add to the action list */
+	a = new_action_list();
+	if (a == NULL)
+	    return SIEVE_NOMEM;
+	b->next = a;
+    }
     a->a = ACTION_FILEINTO;
-    a->cancel_keep = cancel_keep;
+    a->cancel_keep |= cancel_keep;
     a->u.fil.mailbox = mbox;
     a->u.fil.imapflags = imapflags;
-    b->next = a;
-    a->next = NULL;
     return 0;
 }
 
@@ -169,29 +186,46 @@ int do_redirect(action_list_t *a, const char *addr, int cancel_keep)
  *
  * incompatible with: reject
  */
-int do_keep(action_list_t *a, sieve_imapflags_t *imapflags)
+int do_keep(action_list_t *a, int cancel_keep, strarray_t *imapflags)
 {
     action_list_t *b = NULL;
 
     /* see if this conflicts with any previous actions taken on this message */
     while (a != NULL) {
-	b = a;
 	if (a->a == ACTION_REJECT)
 	    return SIEVE_RUN_ERROR;
-	if (a->a == ACTION_KEEP) /* don't bother doing it twice */
-	    return 0;
+	if (a->a == ACTION_KEEP) {
+	    /* don't bother doing it twice */
+	    /* check that we have a valid action */
+	    if (b == NULL) {
+		return SIEVE_INTERNAL_ERROR;
+	    }
+	    /* cut this action out of the list */
+	    b->next = a->next;
+	    a->next = NULL;
+	    /* find the end of the list */
+	    while (b->next != NULL) {
+		b = b-> next;
+	    }
+	    /* add the action to the end of the list */
+	    b->next = a;
+	    break;
+	}
+	b = a;
 	a = a->next;
     }
 
-    /* add to the action list */
-    a = (action_list_t *) xmalloc(sizeof(action_list_t));
-    if (a == NULL)
-	return SIEVE_NOMEM;
+    if(a == NULL) {
+	/* add to the action list */
+	a = new_action_list();
+	if (a == NULL)
+	    return SIEVE_NOMEM;
+	a->next = NULL;
+	b->next = a;
+    }
     a->a = ACTION_KEEP;
-    a->cancel_keep = 1;
+    a->cancel_keep |= cancel_keep;
     a->u.keep.imapflags = imapflags;
-    a->next = NULL;
-    b->next = a;
     return 0;
 }
 
@@ -276,7 +310,7 @@ int do_vacation(action_list_t *a, char *addr, char *fromaddr,
  *
  * incompatible with: reject
  */
-int do_setflag(action_list_t *a, const char *flag)
+int do_setflag(action_list_t *a)
 {
     action_list_t *b = NULL;
  
@@ -294,7 +328,6 @@ int do_setflag(action_list_t *a, const char *flag)
 	return SIEVE_NOMEM;
     a->a = ACTION_SETFLAG;
     a->cancel_keep = 0;
-    a->u.fla.flag = flag;
     b->next = a;
     a->next = NULL;
     return 0;
